/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is WebMail Notifier.
 *
 * The Initial Developer of the Original Code is
 * Byungwook Kang.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

/***********************************************************
constants
***********************************************************/
const ST_CHECK       = 0;
const ST_PRE         = 10;
const ST_PRE_RES     = 11;
const ST_LOGIN       = 20;
const ST_LOGIN_RES   = 21;
const ST_DATA        = 30;
const ST_DATA_RES    = 31;

const LOGIN_URL      = 0;
const LOGIN_ID       = 1;
const LOGIN_PW       = 2;
const LOGIN_DATA     = 3;
const LOGIN_POST     = 1;

/***********************************************************
WebMailHandler definition
***********************************************************/
function WebMailHandler(webMailNotifier){
  this.main=webMailNotifier;
  this.cookieManager=new CookieManager();
  this.cookieManager2=new CookieManager();
  this.loggedIn=false;
  this.reset();
}
WebMailHandler.prototype={
  browser:"firefox",
  charSet: null,
  data: "",
  stage: 0,
  initStage: ST_LOGIN,
  mailCount:-1,
  mailData:{},
  user: "",
  password: null,
  loginData:[],
  started:false,
  retry:1,
  asyncOnChannelRedirect: function(oldChan, newChan, flags, redirectCallback) {
    this.onChannelRedirect(oldChan, newChan, flags);
    redirectCallback.onRedirectVerifyCallback(0);
  },
  // nsIChannelEventSink
  onChannelRedirect: function (aOldChannel, aNewChannel, aFlags) {
    if(this.channel==aOldChannel)this.channel = aNewChannel;
  },
  // nsIInterfaceRequestor
  getInterface: function (aIID) {
    try {
      return this.QueryInterface(aIID);
    } catch (e) {
//dout("getInterface "+aIID);
      throw Components.results.NS_NOINTERFACE;
    }
  },
  // nsIProgressEventSink
  onProgress : function (aRequest, aContext, aProgress, aProgressMax) { },
  onStatus : function (aRequest, aContext, aStatus, aStatusArg) { },
  // nsIHttpEventSink
  onRedirect : function (aOldChannel, aNewChannel) { },
  // nsIAuthPromptProvider
  getAuthPrompt : function(aReason){ return null; },
  QueryInterface : function(aIID) {
    if (aIID.equals(Ci.nsISupports) ||
        aIID.equals(Ci.nsIInterfaceRequestor) ||
        aIID.equals(Ci.nsIChannelEventSink) ||
        aIID.equals(Ci.nsIProgressEventSink) ||
        aIID.equals(Ci.nsIHttpEventSink) ||
        aIID.equals(Ci.nsIStreamListener)||
        aIID.equals(Ci.nsIAuthPromptProvider)||
        aIID.equals(Ci.nsIPrompt)|| //not used but required in ff 2.0
        aIID.equals(Ci.nsIDocShell)|| //required in ff 3.0
        aIID.equals(Ci.nsIAuthPrompt))
      return this;
//dout("QueryInterface "+aIID);
    throw Components.results.NS_NOINTERFACE;
  },
  onStartRequest : function (aRequest, aContext) {
    this.data = "";
  },
  onDataAvailable : function (aRequest, aContext, aStream, aSourceOffset, aLength){
	  var is = Components.classes["@mozilla.org/scriptableinputstream;1"]
            .createInstance(Ci.nsIScriptableInputStream);
  	is.init(aStream);
  	this.data += is.read(aLength);
  },
  onStopRequest : function (aRequest, aContext, aStatus) {
    if(aStatus==Components.results.NS_BINDING_ABORTED)return;
    var charSet=null;
    var httpChannel=aRequest.QueryInterface(Ci.nsIHttpChannel);
    if(this.charSet)charSet=this.charSet;
    else{
      var fnd=this.data.match(/<meta.+?\".+?charset=(\S+?)\s*?;?\s*?\"\s*?\x2f?>/);
      if(fnd)charSet=fnd[1];
      else{
        try{
          if(httpChannel.contentCharset)charSet=httpChannel.contentCharset;
        }catch(e){}
      }
    }
    if(charSet){
      var uniConv = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"]
      								.createInstance(Ci.nsIScriptableUnicodeConverter);
      try{
        uniConv.charset = charSet;
        this.data=uniConv.ConvertToUnicode(this.data);
      }catch(e){}
    }
    try{
      if(!this.process(aContext.QueryInterface(Ci.nsIHttpChannel), this.data))++this.stage;
    }catch(e){
      this.onError();
dout(this.id+" "+e);
    }
  },
  stop : function(){
    if(this.channel){
      this.channel.cancel(Components.results.NS_BINDING_ABORTED);
    }
  },
  process : function (aHttpChannel, aData){
    switch(this.stage){
    case ST_LOGIN:
      this.getHtml(this.loginData[LOGIN_URL],this.loginData[LOGIN_POST]);
      return false;
    case ST_LOGIN_RES:
      this.stage=ST_DATA;
      return this.process(aHttpChannel, aData);
    case ST_DATA:
      if(!this.dataURLCopy)this.dataURLCopy=this.dataURL;
      this.getHtml(this.dataURLCopy);
      return false;
    case ST_DATA_RES:
      var n=parseInt(this.getCount(aData));      
      this.mailCount=isNaN(n)?-1:n;
      this.mailData=this.getData(aData);
      this.mailData.desc=this.getDesc();
      try{
        if(this.mailCount>=0&&aHttpChannel!=null&&aHttpChannel.responseStatus==302){
          /*var l=aHttpChannel.getResponseHeader("Location");
          if(l.indexOf("http")==0){
            this.dataURLCopy=l;
          }else{
            if(l.indexOf("/")==0){
              this.dataURLCopy=aHttpChannel.URI.scheme+"://"+aHttpChannel.URI.hostPort+l;
            }
          }*/
          this.dataURLCopy=this.channel.URI.spec;
        }
      }catch(e){
dout(e);
      }
      if(this.mailCount<0){
        if(this.retry<1){
          ++this.retry;
          this.cookieManager.clear();
          this.checkMail();
          return true;
        }else{
          if(enableDebug!=0)dlog(this.id+" "+this.user+" "+this.stage+"\t"+this.data);
          this.reset();
        }
      }else{
        this.retry=0;
        this.stage=ST_DATA;
      }
      this.main.setState(this.main.MSG_MAILDATA,this.ind);
      if(this.main.autoLogin&&!this.noCookie&&this.main.prefBranch.getBoolPref("autoLoginDefaultAccount")){
        if(this.mailCount>=0&&!this.loggedIn&&this.main.isDefault(this.id,this.user)){
          this.loggedIn=true;
          this.cookieManager.setCookieToBrowser();
        }
      }
      if(this.main.autoLogin&&!this.noCookie){
        if(this.mailCount>=0&&!this.loggedIn){
          if(this.main.multiSession&&(!this.main.isDefault(this.id,this.user)||!this.main.prefBranch.getBoolPref("autoLoginDefaultAccount"))){
            this.loggedIn=true;
            var ar=this.cookieManager.copyTo(this.cookieManager2);
          }
        }
      }      
      return true;
    }
    if(this.stage<ST_PRE&&this.checkLogin){
      return this.checkLogin(aData,aHttpChannel);
    }    
dout("[onError] "+this.id+" "+this.user+" "+this.stage);
    this.onError();
    return true;
  },
  onError : function(){
    if(enableDebug!=0)dlog(this.id+" "+this.user+" "+this.stage+"\t"+this.data);
    this.reset();
    this.main.setState(this.main.MSG_MAILDATA,this.ind);
  },
  reset : function(){
    this.mailCount=-1;
    this.mailData={};
    this.mailData.desc=this.getDesc();
    this.stage=(this.checkLogin?ST_CHECK:this.initStage);
    this.dataURLCopy=null;
    if(!this.checkLogin)this.cookieManager.clear();
  },
  checkMail : function(){
    if(!this.enabled)return;
    this.stop();
    //this.stage=this.mailCount<0?this.initStage:ST_DATA;
    if(this.mailCount<0)this.reset();
    else this.stage=ST_DATA;
    this.started=true;
    try{
      if(!this.process(null,""))++this.stage;
    }catch(e){
      this.onError();
dout(this.id+" "+e);
    }
  },
  onRequest : function(aHttpChannel,second){
    var cm=second?this.cookieManager2:this.cookieManager;
    var ck=cm.getCookie(aHttpChannel.URI,aHttpChannel);
    aHttpChannel.setRequestHeader("Cookie",ck, false);
  },
  onResponse : function(aHttpChannel,second){
    try {
      var cookie = aHttpChannel.getResponseHeader("Set-Cookie");
      var cm=second?this.cookieManager2:this.cookieManager;
      cm.addCookies(aHttpChannel.URI,cookie);
      aHttpChannel.setResponseHeader("Set-Cookie", "", false);
    }catch(e){}
  },
  getMailURL : function(aFolder){
    return this.mailURL;
  },
  getIconURL : function(){
    try{
      var url=this.mailURL.match(/(((\S+):\/\/([^/]+))(\S*\/)?)([^/]*)/);
      if(url)return url[2]+"/favicon.ico";
    }catch(e){}
    return null;
  },
  getIconPage : function(){
    return this.mailURL;
  },
  getCount : function(aData){
    return -1;
  },
  getData : function(aData){
    return {};
  },
  getDesc : function(){
    var n=this.calcCount();
    return n>0?n:"";
  },
  calcCount : function() {
    var aCount=this.mailCount;
    if(this.main&&this.main.resetCounter){
      if(aCount>=0){
        var count=this.loadCount();
        if(aCount>=count)aCount-=count;
        else{
          this.saveCount(aCount>0?aCount:0)
          aCount=0;
        }
      }
    }
    return aCount;
  },
  loadCount:function(){
    var prefStr="accounts."+this.id+"."+this.user+".count";
    var count=null;
    try{
      count=this.main.prefBranch.getIntPref(prefStr);
    }catch(e){}
    if(count==null){
      this.main.prefBranch.setIntPref(prefStr,0);
      count=0;
    }
    return count;
  },
  saveCount:function(n){
    var prefStr="accounts."+this.id+"."+this.user+".count";
    this.main.prefBranch.setIntPref(prefStr,n);
  },
  getForm:function(data,name){
    if(name){
      var reg=new RegExp("<form[\\s\\S]+?name\\s*=\\s*[\"\']"+name+"[\"\']([\\S\\s]+?)<\/form>","i");
      var s=data.match(reg);
      if(!s)return "";
      data=s[1];
    }
    var re=/<input[^>]+?name\s*=\s*[\"\'](\S+?)[\"\'][^>]+?value\s*=s*[\"\']([\s\S]*?)[\"\']/ig;
    var o;
    var post="";
    while ((o = re.exec(data)) != null){
      if(post)post+="&";
      post+=o[1]+"="+encodeURIComponent(o[2]);
    }
    return post;
  },
  delay:function(sec){
    if(this.dTimer)this.dTimer.cancel();
    this.dTimer=Components.classes["@mozilla.org/timer;1"].createInstance(Components.interfaces.nsITimer);
    this.dTimer.initWithCallback(this,sec,Components.interfaces.nsITimer.TYPE_ONE_SHOT); 
  },
  notify:function(aTimer){
    delete this.dTimer;
    if(!this.process(null,""))++this.stage;
  }
};
WebMailHandler.prototype.baseProcess=WebMailHandler.prototype.process
/***********************************************************
load html
***********************************************************/
WebMailHandler.prototype.getHtml = function(aURL,aPostData,aHeaders,aMethod) {
  var ioService = Components.classes["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
  var uri = ioService.newURI(aURL, null, null);
  var channel = ioService.newChannelFromURI(uri);
  var httpChannel = channel.QueryInterface(Ci.nsIHttpChannel);
  if (aPostData||aPostData=="") {
    var uploadStream = Components.classes["@mozilla.org/io/string-input-stream;1"]
                        .createInstance(Ci.nsIStringInputStream);
    uploadStream.setData(aPostData, aPostData.length);
    var uploadChannel = channel.QueryInterface(Ci.nsIUploadChannel);
    uploadChannel.setUploadStream(uploadStream, "application/x-www-form-urlencoded", -1);
    httpChannel.requestMethod = "POST";
  }
  if(aHeaders){
    for(var t in aHeaders){
      httpChannel.setRequestHeader(t,aHeaders[t],false);
    }
  }
  if(aMethod)httpChannel.requestMethod=aMethod;
  this.channel=channel;
  channel.notificationCallbacks = this;
  channel.asyncOpen(this,httpChannel);
}
WebMailHandler.prototype.getCookie = function(domain,name){
  var obj = Components.classes["@mozilla.org/cookiemanager;1"].
              getService(Ci.nsICookieManager2);
  var enm = obj.enumerator;
  while(enm.hasMoreElements()){
    var ck = enm.getNext();
    ck.QueryInterface(Ci.nsICookie2);
    if(endsWith(ck.host,domain)&&ck.name==name)return ck;
  }
  return null;
};
WebMailHandler.prototype.getCookieString = function(domain,name){
  var ck=this.getCookie(domain,name);
  if(!ck)return null;
  var s=ck.name+"="+ck.value;
  if(ck.expires!=0)s+="; expires="+new Date(ck.expires*1000).toUTCString();
  if(ck.path)s+="; path="+ck.path;
  if(ck.host)s+="; domain="+ck.host;
  if(ck.secure)s+="; secure";
  if(ck.httponly)s+="; httponly";
  return s;
};